import wx
from SimulatorFramework.SimApp import *
from version import PolyominoTASVersion
from SimulatorFramework.version import SimAppVersion
import PolyominoTASSystem
import PolyominoTASTileType
import PolyominoTypeEditor
import PolyominoTASAssembly
import PolyominoTASCanvas
from PolyominoTAS3DCanvas import *


SIMFW_ID_Z_PLANE = 201

        
class PolyominoTASApp(SimApp):
    def __init__(self):
        self.dimMode = SIMU_2D_MODE
        SimApp.__init__(self)

    def OnInit(self):
        retVal = SimApp.OnInit(self, title="PolyominoTAS")
        self.tileTypeClassName = PolyominoTASTileType.PolyominoType
        

        # Modify toolbar buttons
        tb_button = self.toolbar.FindById(SIMFW_ID_ENABLEFRONTIER)
        tb_button.SetShortHelp("Step through binding options")
        
        tb_button = self.toolbar.FindById(SIMFW_ID_DISABLEFRONTIER)
        tb_button.SetShortHelp("Select current binding option")

        return retVal
    
    def AddToolbarButtons(self):
        self.toolbar.AddTool(SIMFW_ID_Z_PLANE, "Z plane(s)", wx.Image("images/z.xpm", wx.BITMAP_TYPE_XPM).AdjustChannels(1.0, 1.0, 1.0, 2.0).ConvertToBitmap(), shortHelp="Z plane(s)")

    def OnTool(self, evt):
        if evt.GetId() == SIMFW_ID_Z_PLANE:
            self.canvas.SetZPlanes()
        else:
            SimApp.OnTool(self, evt)

    def OnRefresh(self, event):
        self.canvas.assembly.setDimensions()
        #print(f'minX: {self.canvas.assembly.minX}, maxX: {self.canvas.assembly.maxX}')
        self.canvas.UpdateSizeVars()
        self.canvas.CanvasSizeUpdate(self.canvas.GetCenterCoords())
        self.canvas.Refresh()

    def CreateCanvas(self):
        if self.dimMode == SIMU_2D_MODE:
            self.canvas = PolyominoTASCanvas.PolyominoTASCanvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)
        elif self.dimMode == SIMU_3D_MODE:
            self.canvas = PolyominoTAS3DCanvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)

    def CreateMenu(self):
        menuBar = wx.MenuBar()
        menu = wx.Menu()
        item1 = menu.Append(-1, "&Load system...\tCtrl-L", "Load system...")
        self.Bind(wx.EVT_MENU, self.OnLoad, item1)
        itemLVA = menu.Append(-1, "Load system for viewing only...", "Load system for viewing only...")
        self.Bind(wx.EVT_MENU, self.OnLoadForViewingOnly, itemLVA)
        #itemLA = menu.Append(-1, "Load assembly...\tCtrl-A", "Load assembly...")
        #self.Bind(wx.EVT_MENU, self.OnLoadAssembly, itemLA)
        itemLPCTs = menu.Append(-1, "Load polyomino types from editor", "Load from editor")
        self.Bind(wx.EVT_MENU, self.OnLoadFromEditor, itemLPCTs)
        itemR = menu.Append(-1, "Reload system file", "Reload system from file")
        self.Bind(wx.EVT_MENU, self.OnReload, itemR)
        itemReset = menu.Append(-1, "&Reset to seed\tCtrl-R", "Reset system to seed")
        self.Bind(wx.EVT_MENU, self.OnReset, itemReset)
        #itemSetSeed = menu.Append(-1, "Set seed...\tCtrl-N", "Set seed...")
        #self.Bind(wx.EVT_MENU, self.OnSetSeed, itemSetSeed)
        itemSetAsSeed = menu.Append(-1, "Set current assembly as seed")
        self.Bind(wx.EVT_MENU, self.OnSetAsSeed, itemSetAsSeed)
        item2 = menu.Append(-1, "&Save system as...\tCtrl-S", "Save system as...")
        self.Bind(wx.EVT_MENU, self.OnSaveAs, item2)
        menu.AppendSeparator()
        
        itemSVG = menu.Append(-1, "Save assembly as SVG...", "Save assembly as SVG...")
        self.Bind(wx.EVT_MENU, self.OnSaveAssemblyAsSVG, itemSVG)
        #itemPDF = menu.Append(-1, "Save assembly as PDF...", "Save assembly as PDF...")
        #self.Bind(wx.EVT_MENU, self.OnSaveAssemblyAsPDF, itemPDF)
        menu.AppendSeparator()

        item3 = menu.Append(-1, "E&xit\tCtrl-X", "Exit simulator")
        self.Bind(wx.EVT_MENU, self.OnExitApp, item3)
        menuBar.Append(menu, "&File")

        menuCtrl = wx.Menu()
        
        self.radio2DMode = wx.MenuItem(menuCtrl, -1, text = "2D mode", kind = wx.ITEM_RADIO) 
        self.radio3DMode = wx.MenuItem(menuCtrl, -1, text = "3D mode", kind = wx.ITEM_RADIO)
        menuCtrl.Append(self.radio2DMode)
        menuCtrl.Append(self.radio3DMode)
        self.radio2DMode.Check()
        self.Bind(wx.EVT_MENU, self.OnDimMode, self.radio2DMode)
        self.Bind(wx.EVT_MENU, self.OnDimMode, self.radio3DMode)
        menuCtrl.AppendSeparator()

        itemStep = menuCtrl.Append(-1, "&Forward\tCtrl-F", "Forward one step")
        self.Bind(wx.EVT_MENU, self.OnForward, itemStep)
        itemFF = menuCtrl.Append(-1, "&Go\tCtrl-G", "Fast forward")
        self.Bind(wx.EVT_MENU, self.OnGo, itemFF)
        itemQuit = menuCtrl.Append(-1, "&Quit\tCtrl-Q", "Quit fast forward")
        self.Bind(wx.EVT_MENU, self.OnQuit, itemQuit)
        menuCtrl.AppendSeparator()
        menuBar.Append(menuCtrl, "&Control")

        menuView = wx.Menu()
        itemRefreshRate = menuView.Append(-1, "&Refresh screen every...", "Set screen refresh rate")
        self.Bind(wx.EVT_MENU, self.OnSetRefreshRate, itemRefreshRate)
        menuView.AppendSeparator()
        itemSTC = menuView.Append(-1, "&Center\tCtrl+C", "Scroll to center")
        self.Bind(wx.EVT_MENU, self.OnCenter, itemSTC)
        itemCOC = menuView.Append(-1, "&Center on coordinates...", "Scroll coordinates to center")
        self.Bind(wx.EVT_MENU, self.OnCenterOnCoords, itemCOC)
        menuView.AppendSeparator()
        itemZIn = menuView.Append(-1, "Zoom in\tCtrl-+", "Zoom in")
        self.Bind(wx.EVT_MENU, self.OnZoomIn, itemZIn)
        itemZOut = menuView.Append(-1, "Zoom out\tCtrl--", "Zoom out")
        self.Bind(wx.EVT_MENU, self.OnZoomOut, itemZOut)
        menuView.AppendSeparator()
        itemTG = menuView.Append(-1, "&Toggle grid lines\tCtrl-T", "Toggle display of grid lines")
        self.Bind(wx.EVT_MENU, self.OnToggleGrid, itemTG)
        itemTD = menuView.Append(-1, "&Toggle display updates\tCtrl-D", "Toggle display of tile additions")
        self.Bind(wx.EVT_MENU, self.OnToggleDisplay, itemTD)
        #itemSTT = menuView.Append(-1, "Show only matching tile types...", "Set patterns for tile type names to display")
        #self.Bind(wx.EVT_MENU, self.OnShowMatching, itemSTT)

        #=======================================================================
        # menuView.AppendSeparator()
        # self.radioShowBitmaps = wx.MenuItem(menuView, -1, text = "Show only bitmaps",kind = wx.ITEM_RADIO) 
        # self.radioShowSignals = wx.MenuItem(menuView, -1, text = "Show only signals",kind = wx.ITEM_RADIO)
        # self.radioShowBoth = wx.MenuItem(menuView, -1, text = "Show bitmaps and signals",kind = wx.ITEM_RADIO)  
        # menuView.Append(self.radioShowBitmaps)
        # menuView.Append(self.radioShowSignals)
        # menuView.Append(self.radioShowBoth)
        # self.radioShowBitmaps.Check()
        # self.Bind(wx.EVT_MENU, self.OnShowBitmapsOrLines, self.radioShowBitmaps)
        # self.Bind(wx.EVT_MENU, self.OnShowBitmapsOrLines, self.radioShowSignals)
        # self.Bind(wx.EVT_MENU, self.OnShowBitmapsOrLines, self.radioShowBoth)
        #=======================================================================
        
        menuView.AppendSeparator()
        itemEditorWin = menuView.Append(-1, "Show Polyomino Type editor", "Show Polyomino Type editor")
        self.Bind(wx.EVT_MENU, self.OnToggleViewEditor, itemEditorWin)
        
        menuView.AppendSeparator()
        itemRefresh = menuView.Append(-1, "Refresh screen\tF2", "Refresh screen")
        self.Bind(wx.EVT_MENU, self.OnRefresh, itemRefresh)

        menuBar.Append(menuView, "&View")

        menu2 = wx.Menu()
        itemAbout = menu2.Append(-1, "About", "About")
        self.Bind(wx.EVT_MENU, self.OnAbout, itemAbout)
        itemH = menu2.Append(-1, "&Help\tCtrl-H", "Show help")
        self.Bind(wx.EVT_MENU, self.OnHelp, itemH)
        menuBar.Append(menu2, "&Help")
        
        self.frame.SetMenuBar(menuBar)
        
    def OnSetRefreshRate(self, event):
        doAgain = True
        refresh_rate = self.canvas.refreshRate
        while doAgain == True:
            doAgain = False
            dlg = wx.TextEntryDialog(self.frame, "Enter the number of simulation steps between screen refreshes:", caption="Enter refresh rae", value=str(refresh_rate))
            if not dlg.ShowModal() == wx.ID_OK:
                return
            result = dlg.GetValue()
            dlg.Destroy()
            try:
                refresh_rate = int(result)
            except:
                doAgain = True
        self.canvas.refreshRate = refresh_rate

    def OnToggleViewEditor(self, event):
        self.bShowEditor = not self.bShowEditor
        if self.bShowEditor == True:
            if self.editorFrame is None:
                self.editorFrame = PolyominoTypeEditor.PolyominoTypeEditorFrame(self, self.tileTypeClassName)
            else:
                self.editorFrame.Show(True)
            self.editorFrame.canvas.SetTileTypes(self.canvas.polyominoSystem.PolyominoType_dict)
        elif not self.editorFrame is None:
            self.editorFrame.OnCloseFrame(event)

    def EditorFrameClosed(self):
        self.editorFrame = None
        self.bShowEditor = False
        
    def OnLoadForViewingOnly(self, evt):
        self.canvas.DoLoad(fileName=None, b_viewing_only=True)
        self.canvas.UpdateSizeVars()
        self.canvas.CanvasSizeUpdate()

    def OnLoadFromEditor(self, event):
        if self.editorFrame is None:
            SimUtils.globalApp.ShowErrorDialog("Polyomino Editor has not yet been opened")
            return
        
        # See if there are any changed types
        #    -> if so, ask about throwing out current assembly
        #    -> if not, carefully load new set...
        # set from editor must be a super set
        simPT_dict = self.canvas.polyominoSystem.PolyominoType_dict
        edPT_list = self.editorFrame.canvas.tileTypes
        
        bFoundAll = False
#         bFoundAll = True
#         for ass_loc in self.canvas.assembly.locationMap:
#             ass_polyomino = self.canvas.assembly.locationMap[ass_loc]
#             ass_pt_name = ass_polyomino.polyominoType.name
#             sim_pt = simPT_dict[ass_pt_name]
#             bFound = False
#             for ed_pt in edPT_list:
#                 if ed_pt == sim_pt:
#                     bFound = True
#                     break
#             if bFound is False:
#                 self.OutputMessage(f"Warning: Did not find identical Polyomino type '{ass_pt_name}' in the editor's set")
#                 bFoundAll = False
        if not bFoundAll:
            dlg = wx.MessageDialog(self.frame, "Do you wish to delete the current assembly to continue?", "Confirm assembly delete", style=wx.OK|wx.CANCEL|wx.CENTRE|wx.ICON_WARNING)
            #dlg = wx.MessageDialog(self.frame, "One or more polyominoes in the assembly have types\nwhich are not (the same) in the editor's set.\nDo you wish to delete the current assembly to continue?", "Confirm assembly delete", style=wx.OK|wx.CANCEL|wx.CENTRE|wx.ICON_WARNING)
            if not dlg.ShowModal() == wx.ID_OK:
                dlg.Destroy()
                return
            else:
                dlg.Destroy()
                self.canvas.assembly = PolyominoTASAssembly.PolyominoAssembly()
                self.canvas.seed = self.canvas.assembly
        simPT_dict = dict()
        for ed_pt in edPT_list:
            simPT_dict[ed_pt.name] = ed_pt
        self.OutputMessage("Successfully loaded Polyomino types from the editor")
        self.canvas.Refresh()
            

    def OnShowBitmapsOrLines(self, event):
        if self.radioShowSignals.IsChecked():
            self.canvas.drawBitmaps = False
            self.canvas.drawSignals = True
        elif self.radioShowBitmaps.IsChecked():
            self.canvas.drawBitmaps = True
            self.canvas.drawSignals = False
        else:
            self.canvas.drawBitmaps = True
            self.canvas.drawSignals = True 
        self.canvas.Refresh()
        
    def OnSetSeed(self, event):
        self.canvas.DoSetSeed()
        
    def OnSaveSAAAs(self, event):
        self.canvas.DoSaveSAAssemblyAs()
        
    def GetVersionString(self):
        return PolyominoTASVersion + " (SimApp version " + SimAppVersion + ")"

    def OnAbout(self, event):
        versionString = self.GetVersionString()
        aboutInfo = wx.adv.AboutDialogInfo()
        aboutInfo.SetName("PolyominoTAS")
        aboutInfo.SetVersion(versionString)
        aboutInfo.SetDescription("Polyomino Tiling Simulator")
        #aboutInfo.SetCopyright("(C) 2017-2019")
        aboutInfo.SetWebSite("http://www.self-assembly.net")
        aboutInfo.AddDeveloper("Matthew Patitz")
        aboutInfo.AddDeveloper("Daniel Hader")
        wx.adv.AboutBox(aboutInfo)
        
    def SetMouseMode(self, mode):
        self.canvas.mouseMode = mode
        if mode == SimCanvas.SIMFW_MODE_PAN_ZOOM:
            self.canvas.rubberBand.reset()
            self.canvas.SetCursor(wx.Cursor(self.canvas.grabCursorImg))
            self.toolbar.EnableTool(SIMFW_ID_COPY, False)
            self.toolbar.EnableTool(SIMFW_ID_PASTE, False)
            self.toolbar.EnableTool(SIMFW_ID_DISABLEFRONTIER, False)
            self.toolbar.EnableTool(SIMFW_ID_ENABLEFRONTIER, False)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_IN, False)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_OUT, False)
            self.toolbar.EnableTool(SIMFW_ID_UNDO, False)
        else:
            self.canvas.currentPolyominoType = PolyominoTASTileType.PolyominoType(name="New type")
            self.canvas.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
            self.toolbar.EnableTool(SIMFW_ID_COPY, True)
            self.toolbar.EnableTool(SIMFW_ID_PASTE, True)
            self.toolbar.EnableTool(SIMFW_ID_DISABLEFRONTIER, True)
            self.toolbar.EnableTool(SIMFW_ID_ENABLEFRONTIER, True)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_IN, True)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_OUT, True)
            self.toolbar.EnableTool(SIMFW_ID_UNDO, True)

    def OnDimMode(self, event):
        if self.radio2DMode.IsChecked():
            self.SetDimMode(SIMU_2D_MODE)
        elif self.radio3DMode.IsChecked():
            self.SetDimMode(SIMU_3D_MODE)
        
    def SetDimMode(self, mode):
        if mode == self.dimMode:    # No change
            return
        if (not mode == SIMU_2D_MODE) and (not mode == SIMU_3D_MODE):
            SimUtils.globalApp.OutputMessage("Invalid dimension mode requested: " + str(mode))
            return

        #NOTE: creation of a new canvas re-seeds the random number generator, which may not be desired
        self.dimMode = mode
        
        if self.dimMode == SIMU_2D_MODE:
            self.radio2DMode.Check(True)
            #self.SetCaching(True)
            #self.checkKeepCache.Enable(True)
            #self.itemOBJ.Enable(False)
        else:
            self.radio3DMode.Check(True)
            #self.SetCaching(False)
            #self.checkKeepCache.Enable(False)
            #self.itemOBJ.Enable(True)

        pane = self.frame._mgr.GetPane(self.canvas) 
        if pane.IsOk():
            self.canvas.UnsubscribeEventHandlers()
            #self.frame._mgr.ClosePane(pane) 
            self.frame._mgr.DetachPane(pane.window) 
            pane.window.Destroy() 

        self.CreateCanvas()
        self.frame._mgr.AddPane(self.canvas, wx.CENTER, 'Simulation Window')
        self.canvas.SubscribeEventHandlers()
        
        #if self.dimMode == SIMU_3D_MODE:
        #    self.CreateTilesPerStepWin()
        #else:
        #    self.frame._mgr.DetachPane()

        self.frame._mgr.Update()

    def OnSaveAssemblyAsSVG(self, evt: wx.Event):
        saveFileDialog = wx.FileDialog(self.frame, "Save assembly to SVG as...", "", "", 
                                       "SVG files (*.svg)|*.svg",
                                       wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        saveFileDialog.ShowModal()
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        if len(fileName) > 0:
            # MJP TODO: Ask if frontier locations should be included
            # MJP TODO: Ask if the PDF should show the sequence

            self.canvas.assembly.WriteToPDForSVG(svgFileName=fileName, asSequence=False, displayMask=None, 
                                                 sideLen=self.canvas.sideLen, z_range=self.canvas.z_planes_display)
        
    def OnSaveAssemblyAsPDF(self, evt: wx.Event):
        saveFileDialog = wx.FileDialog(self.frame, "Save assembly to PDF as...", "", "", 
                                       "PDF files (*.pdf)|*.pdf",
                                       wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        saveFileDialog.ShowModal()
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        if len(fileName) > 0:
            # MJP TODO: Ask if frontier locations should be included
            # MJP TODO: Ask if the PDF should show the sequence

            self.canvas.assembly.WriteToPDForSVG(pdfFileName=fileName, asSequence=False, displayMask=None,
                                                 sideLen=self.canvas.sideLen, z_range=self.canvas.z_planes_display)

#----------------------------------------------------------------------------
# The following lines create the actual application and begin its execution
app = PolyominoTASApp()
app.MainLoop()

#---------------------------------------------------------------------------

