import time
from SimulatorFramework.TileAssembly import *
from SimulatorFramework.TileType import Tile
from SimulatorFramework.TileType import TileType
from SimulatorFramework.TileType import *
from SimulatorFramework import SimUtils
import os
import ast
import random
import itertools
from PolyominoTASTileType import Polyomino, PolyominoType, DIR_OFFSETS, labels_complement
from StochasticEvents import *
import svgwrite
from svgwrite import cm, mm, px
from svgwrite import pt as svgPt
from reportlab.pdfgen import canvas


NBRS = [(0,1,0),(1,0,0),(0,-1,0),(-1,0,0),(0,0,1),(0,0,-1)]
#==============================
# N = 0
# E = 1
# S = 2
# W = 3
# U = 4
# D = 5
#=============================#
OPP_DIR = {0:2, 1:3, 2:0, 3:1, 4:5, 5:4}

EMPTY_LOC           = 0
FRONTIER_LOC        = 1
FRONTIER_LOC_OFF    = 2


class PolyominoAssembly:
    def __init__(self, polyomino_system = None):
        self.NBRS = [(0,1,0),(1,0,0),(0,-1,0),(-1,0,0),(0,0,1),(0,0,-1)]
        self.DIRSTRS = ["N", "E", "S", "W", "U", "D"]
        self.OPPDIRS = [2, 3, 0, 1, 5, 4]

        self.growthInc = 20
        self.tau = 2
        self.num_polyominoes = 0
        self.locationMap = {} # map associating locations to polyomino and component

        # Data structures to keep track of the locations where polyominoes
        # can be added in the kinetic model
        self.attachable_polyominoes = {}
        self.attachment_event_map = {}
        
        # Data structure to keep track of binding strengths of attached polyominoes
        self.binding_strength_map = {}
        
        # Hack to make kinetic model simulations faster when set > 0
        self.min_binding = None
        
        self.frontier = {}
        #TODO: to speed things up, keep a fronterList which is a list of the frontier locations
        #      for quick random selection. Currently will have to use random.choice(list(self.frontier.values()))

        self.minX = -20
        self.maxX =  20
        self.minY = -20
        self.maxY =  20
        self.minZ =  0
        self.maxZ =  0

        self.systemDefFile = None
        self.polyominoSystem = polyomino_system
        self.b_check_for_local_determinism = True

    def ClearSpace(self):
        del self.locationMap
        del self.frontier
        self.num_polyominoes = 0
        self.locationMap = {}
        self.frontier = {}
        
    def GetWidth(self):
        #self.setDimensions()
        return self.width

    def GetHeight(self):
        #self.setDimensions()
        return self.height
    
    def GetDepth(self):
        #self.setDimensions()
        return self.depth

    def GetNumTiles(self):
        return self.num_polyominoes
    
    #===========================================================================
    # def GetPolyominoes(self):
    #     polyominoes = []
    #     for loc in self.locationMap:
    #         polyomino = self.locationMap[loc]
    #         if not polyomino in polyominoes:
    #             polyominoes.append(polyomino)
    #     return polyominoes
    #===========================================================================
    
    def setDimensions(self):
        ((minX, minY, minZ), (maxX, maxY, maxZ)) = self.getBoundingBox()
        self.minX = minX - 5
        self.maxX = maxX + 5
        self.minY = minY - 5
        self.maxY = maxY + 5
        self.minZ = minZ
        self.maxZ = maxZ
        
        #=======================================================================
        # if minX <= self.minX:
        #     self.minX = self.minX - self.growthInc
        # if maxX >= self.maxX:
        #     self.maxX = self.maxX + self.growthInc
        # if minY <= self.minY:
        #     self.minY = self.minY - self.growthInc
        # if maxY >= self.maxY:
        #     self.maxY = self.maxY + self.growthInc
        # if minZ <= self.minZ:
        #     self.minZ = self.minZ - 1#self.growthInc
        # if maxZ >= self.maxZ:
        #     self.maxZ = self.maxZ + 1#self.growthInc
        #=======================================================================
        self.width  = self.maxX - self.minX
        self.height = self.maxY - self.minY
        self.depth  = self.maxZ - self.minZ
        self.numTiles = self.GetNumTiles()
        #print(f'in setDimensions, boundingbox={self.getBoundingBox()}')
        #print(f'in setDimensions, width={self.width}, height={self.height}, minX={self.minX}, maxX={self.maxX}, minY={self.minY}, maxY={self.maxY}, minZ={self.minZ}, maxZ={self.maxZ}')

    def getBoundingBox(self):
        min_x = min_y = min_z = 999999
        max_x = max_y = max_z = -999999

        if len(self.locationMap) > 0:            
            for loc in self.locationMap:
                if not self.locationMap[loc] is None:
                    if loc[0] < min_x:
                        min_x = loc[0]
                    if loc[1] < min_y:
                        min_y = loc[1]
                    if loc[2] < min_z:
                        min_z = loc[2]
                    if loc[0] > max_x:
                        max_x = loc[0]
                    if loc[1] > max_y:
                        max_y = loc[1]
                    if loc[2] > max_z:
                        max_z = loc[2]

        return ((min_x, min_y, min_z), (max_x, max_y, max_z))

    def OverlapsExistingPolyomino(self, polyomino):
        for block in polyomino.blocks:
            if block in self.locationMap:
                return True
        return False

    def PrintDebugInfo(self):
        SimUtils.globalApp.OutputMessage(f"Not currently implemented.")
        return
        self.PrintAssemblyStatistics()

        SimUtils.globalApp.OutputMessage(f"[DEBUG] Polyominoes in the assembly:")
        for domain_loc in self.locationMap:
            existing_polyomino = self.locationMap[domain_loc]
            SimUtils.globalApp.OutputMessage(f'  {existing_polyomino}')
        
        SimUtils.globalApp.OutputMessage("[DEBUG] Attachable polyominoes:")
        for polyomino_hash in self.attachable_polyominoes:
            attachable_polyomino = self.attachable_polyominoes[polyomino_hash]
            SimUtils.globalApp.OutputMessage(f'  hash: {polyomino_hash}, polyomino: {attachable_polyomino}')
        
    def GetBindingStrength(self, polyomino):
        binding_strength = 0
        for domain_loc in polyomino.bindable_nbr_glues:
            if domain_loc in self.locationMap:
                existing_polyomino = self.locationMap[domain_loc]
                if domain_loc in existing_polyomino.domains:
                    existing_domains_at_loc = existing_polyomino.domains[domain_loc]
                    for existing_domain in existing_domains_at_loc:
                        if existing_domain == polyomino.bindable_nbr_glues[domain_loc]:
                            binding_strength += existing_domain[2]
        return binding_strength

    def RemoveAttachmentEventFromMap(self, coords, polyomino_hash):
        self.attachment_event_map[coords].remove(polyomino_hash)
        if len(self.attachment_event_map[coords]) == 0:
            del self.attachment_event_map[coords]

    def RemoveAttachmentEventForPolyomino(self, polyomino):
        polyomino_hash = polyomino.hash
        for block in polyomino.blocks:
            self.RemoveAttachmentEventFromMap(block, polyomino_hash)
        del self.attachable_polyominoes[polyomino_hash]
        self.polyominoSystem.stochastic_event_set.remove_attachment_event(polyomino.get_reference_point_loc(), polyomino)
            
    def RemoveAttachmentEventsForLocation(self, coords):
        # remove any events that may have already existed for the locations of the new polyomino
        if coords in self.attachment_event_map:
            for polyomino_hash in self.attachment_event_map[coords]:
                polyomino = self.attachable_polyominoes[polyomino_hash]
                for block in polyomino.blocks:
                    if block != coords:
                        self.RemoveAttachmentEventFromMap(block, polyomino_hash)

                del self.attachable_polyominoes[polyomino_hash]
                self.polyominoSystem.stochastic_event_set.remove_attachment_event(polyomino.get_reference_point_loc(), polyomino)
            del self.attachment_event_map[coords]
                

    # update the binding strengths of any polyominoes to which this may have bound if not a seed polyomino
    def UpdateBindingStrength(self, polyomino):
        polyomino_hash = polyomino.hash
        if polyomino_hash in self.binding_strength_map:
            binding_strength = self.GetBindingStrength(polyomino)
            if not binding_strength == self.binding_strength_map[polyomino_hash]:
                ref_loc = polyomino.get_reference_point_loc()
                self.polyominoSystem.stochastic_event_set.remove_detachment_event(ref_loc, self.binding_strength_map[polyomino_hash])
                self.binding_strength_map[polyomino_hash] = binding_strength
                self.polyominoSystem.stochastic_event_set.add_detachment_event(ref_loc, binding_strength)    

    def PrintAssemblyStatistics(self):
        polyominoes = self.GetPolyominoes()
        num_polyominoes = len(polyominoes)

        mismatched_domains_count = 0
        half_mismatched_domains_count = 0
        unbound_domains_count = 0
        strength_dict = dict()
        
        for polyomino in polyominoes:
            binding_strength = 0
            for domain_loc in polyomino.bindable_nbr_glues:
                if domain_loc in self.locationMap:
                    existing_polyomino = self.locationMap[domain_loc]
                    b_mismatch = False
                    if domain_loc in existing_polyomino.domains:
                        b_mismatch = True
                        existing_domains_at_loc = existing_polyomino.domains[domain_loc]
                        for existing_domain in existing_domains_at_loc:
                            if existing_domain == polyomino.bindable_nbr_glues[domain_loc]:
                                binding_strength += existing_domain[2]
                                b_mismatch = False
                    else:
                        half_mismatched_domains_count += 1
                    if b_mismatch == True:
                        mismatched_domains_count += 1
                else:
                    unbound_domains_count += 1
            
            # Keep track of the binding strengths of polyominoes
            if binding_strength not in strength_dict:
                strength_dict[binding_strength] = []
            strength_dict[binding_strength].append(polyomino)
            
        # Mismatched domains will have been counted on both sides of a mismatch
        mismatched_domains_count = int(mismatched_domains_count / 2)
                        
        # Print out stats
        print(f'Statistics of the assembly:')
        print(f'  Number of polyominoes: {num_polyominoes}')
        print(f'  Number of mismatched domain pairs: {mismatched_domains_count}')
        print(f'  Number of domains adjacent to blocks without domains: {half_mismatched_domains_count}')
        print(f'  Number of domains adjacent to empty space: {unbound_domains_count}')
        for strength in strength_dict:
            print(f'    Number of polyominoes bound by strength {strength}: {len(strength_dict[strength])}')        
                            
    def AddPolyominoKineticModel(self, polyomino, b_seed=False):
        #print(f'DEBUG Adding polyomino {polyomino}')
        if self.OverlapsExistingPolyomino(polyomino):
            SimUtils.globalApp.OutputMessage(f"Error: Attempted to add polyomino '{polyomino}' which overlaps a portion of the assembly")
            return False

        # create an event for the detachment of this polyomino based on its binding strength if this isn't a seed polyomino
        if not b_seed:
            binding_strength = self.GetBindingStrength(polyomino)
            ref_coords = polyomino.get_reference_point_loc()
            self.polyominoSystem.stochastic_event_set.add_detachment_event(ref_coords, binding_strength)
            self.binding_strength_map[polyomino.hash] = binding_strength

        for block in polyomino.blocks:
            # add each block to the location map
            self.locationMap[block] = polyomino
            # remove attachment events for potential polyominoes that overlapped this location
            self.RemoveAttachmentEventsForLocation(block)
        self.num_polyominoes = self.num_polyominoes + 1
            
        test_locations = []
        perimeter_locs = polyomino.getPerimeterLocations()
        for pl in perimeter_locs:
            if pl not in self.locationMap:
                test_locations.append(pl)
            else:
                # update binding strengths and associated events of any bound neighbors if changed
                nbr_polyomino = self.locationMap[pl]
                self.UpdateBindingStrength(nbr_polyomino)
                
        possible_additions = self.GetFittingPolyomioesForLocations(test_locations)
        for pa_loc in possible_additions:
            for p in possible_additions[pa_loc]:
                p_hash = p.hash
                p_loc = p.get_reference_point_loc()
                
                if p_hash not in self.attachable_polyominoes:
                    # add an event for the attachment of a polyomino of this type here
                    self.polyominoSystem.stochastic_event_set.add_attachment_event(p_loc, p)
                    for p_block in p.blocks:
                        if p_block in self.attachment_event_map:
                            self.attachment_event_map[p_block].append(p_hash)
                        else:
                            self.attachment_event_map[p_block] = [p_hash]
                    self.attachable_polyominoes[p_hash] = p
                    
    def RemovePolyominoKineticModel(self, coords):
        if not coords in self.locationMap:
            SimUtils.globalApp.OutputMessage(f"Error: Attempted to remove nonexistent polyomino from location {coords}")
            return False
        polyomino = self.locationMap[coords]

        # remove the detachment event from the set of possible events since it just happened
        binding_strength = self.GetBindingStrength(polyomino)
        self.polyominoSystem.stochastic_event_set.remove_detachment_event(polyomino.get_reference_point_loc(), binding_strength)
        del self.binding_strength_map[polyomino.hash]

        for block in polyomino.blocks:
            self.locationMap.pop(block)
        self.num_polyominoes = self.num_polyominoes - 1

        perimeter_locs = polyomino.getPerimeterLocations()
        to_remove = set()
        for pl in perimeter_locs:
            if pl in self.locationMap:
                # update binding strengths and associated events of any bound neighbors if changed
                nbr_polyomino = self.locationMap[pl]
                self.UpdateBindingStrength(nbr_polyomino)
 
            else:
                # remove 'orphaned' events (i.e. those that are now too far away from the perimeter)
                if pl in self.attachment_event_map:
                    for polyomino_hash in self.attachment_event_map[pl]:
                        nbr_polyomino = self.attachable_polyominoes[polyomino_hash]
                        # see if this nbr is adjacent to any polyomino
                        nbr_perimeter_locs = nbr_polyomino.getPerimeterLocations()
                        b_keep = False
                        for npl in nbr_perimeter_locs:
                            if npl in self.locationMap:
                                b_keep = True
                                break
                        if not b_keep:
                            to_remove.add(polyomino_hash)
        for remove_hash in to_remove:
            pol_remove = self.attachable_polyominoes[remove_hash]
            self.RemoveAttachmentEventForPolyomino(pol_remove)
            
        # add events for any possible attachments in this location
        possible_additions = self.GetFittingPolyomioesForLocations(polyomino.blocks)
        for pa_loc in possible_additions:
            for p in possible_additions[pa_loc]:

                # check to make sure this polyomino is actually adjacent to assembly
                # it's possible that it's not
                nbr_locs = p.getPerimeterLocations()
                b_add = False
                for nbr_loc in nbr_locs:
                    if nbr_loc in self.locationMap:
                        b_add = True
                        break
                    
                if not b_add:
                    continue
                
                # add an event for the attachment of a polyomino of this type here
                self.polyominoSystem.stochastic_event_set.add_attachment_event(p.get_reference_point_loc(), p)
                p_hash = p.hash
                self.attachable_polyominoes[p_hash] = p
                for p_block in p.blocks:
                    if p_block in self.attachment_event_map:
                        self.attachment_event_map[p_block].append(p_hash)
                    else:
                        self.attachment_event_map[p_block] = [p_hash]
        
    def GetFittingPolyomioesForLocations(self, location_list):
        possible_additions = {}
        for loc in location_list:
            # try to place every (rotated) polyomino into this location and determine how many bonds it could form
            for pct_name in self.polyominoSystem.PolyominoType_dict:
                pct = self.polyominoSystem.PolyominoType_dict[pct_name]
                
                # make a polyomino for each translation here, i.e. try placing each block of this polyominoType in this location and see if it fits
                for block in pct.blocks:
                    offset = (loc[0]-block[0], loc[1]-block[1], loc[2]-block[2])
                    test_polyomino = Polyomino(pct)
                    test_polyomino.translate(offset)
                    
                    if not self.OverlapsExistingPolyomino(test_polyomino):
                        if (self.min_binding is None) or (self.GetBindingStrength(test_polyomino) >= self.min_binding):
                            ref_point_loc = test_polyomino.get_reference_point_loc()
                            if not ref_point_loc in possible_additions:
                                possible_additions[ref_point_loc] = [test_polyomino]
                            elif not test_polyomino in possible_additions[ref_point_loc]:
                                possible_additions[ref_point_loc].append(test_polyomino)
        return possible_additions

    def AddPolyominoForViewingOnly(self, polyomino):
        for block in polyomino.blocks:
            self.locationMap[block] = polyomino
        self.num_polyominoes += 1
        #self.setDimensions()

        
    def AddPolyomino(self, polyomino):
        if self.OverlapsExistingPolyomino(polyomino):
            SimUtils.globalApp.OutputMessage(f"Error: Attempted to add polyomino '{polyomino}' which overlaps a portion of the assembly")
            return False

        # add to location maps and remove frontier locations that are now occupied
        for block in polyomino.blocks:
            self.locationMap[block] = polyomino
            if block in self.frontier:
                self.frontier.pop(block)

        # compute new frontier locations
        for block in polyomino.blocks:
            if block in polyomino.domains:  # see if it has a glue
                domains = polyomino.domains[block]
                for xo in range(-1, 2):
                    for yo in range(-1, 2):
                        for zo in range(-1, 2):
                            if abs(xo) + abs(yo) + abs(zo) != 1:
                                continue
                            nblock = (block[0] + xo, block[1] + yo, block[2] + zo)
                            for domain in domains:
                                do = (DIR_OFFSETS[domain[1]][0], DIR_OFFSETS[domain[1]][1], DIR_OFFSETS[domain[1]][2])
                                if do == (xo,yo,zo):
                                    oppdir = (-1*do[0], -1*do[1], -1*do[2])
                                    if nblock in self.frontier:
                                        self.frontier[nblock].append((oppdir, domain[0], domain[2]))
                                    else:
                                        self.frontier[nblock] = [(oppdir, domain[0], domain[2])]
                         
        self.num_polyominoes += 1           
        self.setDimensions()
        #print(f'locationmap: {self.locationMap}')
        #for floc in self.frontier:
        #    nbrGlues = self.GetNeighboringGlues(floc)
        #    print(f'floc {floc} has nbrGlues {nbrGlues}')
        #print(f'loc map afer add: {self.locationMap}')

    def RemovePolyomino(self, polyomino):
        oldBlocks = []
        for block in polyomino.blocks:
            self.locationMap.pop(block)
            #  If the location of a removed block has neighboring glues, add it to the frontier
            nbrGlues = self.GetNeighboringGlues(block)
            if len(nbrGlues) > 0:
                self.frontier[block] = nbrGlues
            oldBlocks.append(block)

        #  Remove neighboring locations from the frontier if they have no other incident glues
        for block in oldBlocks:
            for xo in range(-1, 2):
                for yo in range(-1, 2):
                    for zo in range(-1, 2):
                        if abs(xo) + abs(yo) + abs(zo) != 1:
                            continue
                        nblock = (block[0] + xo, block[1] + yo, block[2] + zo)
                        if nblock in self.frontier:
                            nbrGlues = self.GetNeighboringGlues(nblock)
                            if len(nbrGlues) == 0:
                                self.frontier.pop(nblock)
                            else:
                                self.frontier[nblock] = nbrGlues
        self.num_polyominoes -= 1
        self.setDimensions()
        
    def GetNeighboringGlues(self, block):
        nbrGlues = []
        for xo in range(-1, 2):
            for yo in range(-1, 2):
                for zo in range(-1, 2):
                    if abs(xo) + abs(yo) + abs(zo) != 1:
                        continue
                    nloc = (block[0] + xo, block[1] + yo, block[2] + zo)
                    if nloc in self.locationMap:
                        npol = self.locationMap[nloc]
                        if nloc in npol.domains:
                            for domain in npol.domains[nloc]:
                                do = (DIR_OFFSETS[domain[1]][0], DIR_OFFSETS[domain[1]][1], DIR_OFFSETS[domain[1]][2])
                                oppdir = (-1*do[0], -1*do[1], -1*do[2])
                                if oppdir == (xo,yo,zo):
                                    nbrGlues.append(((xo,yo,zo), domain[0], domain[2]))
        return nbrGlues
    
    def GetLoc(self, coords):
        if coords in self.locationMap:
            polyomino = self.locationMap[coords]
            return polyomino
        else:
            return None

    def GetBlockBindingInFrontierLocation(self, frontier_loc, domains):
        #print(f' for frontier loc {loc} {floc}, try domains {pct.domains[dloc]}')
        binding_strength = 0
        for adj_glue in frontier_loc:
            for domain in domains:
                do = (DIR_OFFSETS[domain[1]][0], DIR_OFFSETS[domain[1]][1], DIR_OFFSETS[domain[1]][2])
                if (do == adj_glue[0]) and (labels_complement(domain[0], adj_glue[1]) == True) and adj_glue[2] == domain[2]:
                    binding_strength += domain[2]
        return binding_strength

    def StepForwardKinetic(self):
        random_event = self.polyominoSystem.stochastic_event_set.sample_event(True) # TODO: should the arg be True or False?
        if isinstance(random_event, AttachmentEvent):
            self.AddPolyominoKineticModel(random_event.polyomino)
        elif isinstance(random_event, DetachmentEvent):
            self.RemovePolyominoKineticModel(random_event.coords)
            
        return True # Just returning True for now to force redraw
        
    def StepForward(self):
        if len(self.frontier) == 0:
            return None

        frontier_loc_list = list(self.frontier.keys())
        start_loc = random.randrange(len(frontier_loc_list))
        
        attachable_polyominoes = dict()
        for i in range(len(frontier_loc_list)):
            loc = frontier_loc_list[(start_loc+i)%len(frontier_loc_list)]
            floc = self.frontier[loc]
            for nbr_domain in floc:
            
                self.temp_polyominoes = []
                num_options = 0
                for pct_name in self.polyominoSystem.PolyominoType_dict:
                    pct = self.polyominoSystem.PolyominoType_dict[pct_name]
                    
                    for dloc in pct.domains:
                        binding_strength = self.GetBlockBindingInFrontierLocation(floc, pct.domains[dloc])
                        if binding_strength > 0:
                            trans = (0-dloc[0]+loc[0], 0-dloc[1]+loc[1], 0-dloc[2]+loc[2])
                            new_polyomino = Polyomino(pct)
                            new_polyomino.translate(trans)
                            if self.OverlapsExistingPolyomino(new_polyomino) == False:
                                # polyomino can be placed here and has at least one glue match
                                binding_strength = 1
                                if self.polyominoSystem.binding_threshold > 1:
                                    binding_strength = 0
                                    for domain_coords in new_polyomino.domains:
                                        if domain_coords in self.frontier:
                                            floc2 = self.frontier[domain_coords]
                                            domains2 = new_polyomino.domains[domain_coords]
                                            binding_strength += self.GetBlockBindingInFrontierLocation(floc2, domains2)
                                if binding_strength >= self.polyominoSystem.binding_threshold:
                                    attachable_polyominoes[new_polyomino.hash] = new_polyomino
                                    
    
        if len(attachable_polyominoes) > 0:
            if SimUtils.globalApp.canvas.refreshRate == 0:
                SimUtils.globalApp.OutputMessage(f"{len(attachable_polyominoes)} polyomino attachments were possible")
            if self.b_check_for_local_determinism:
                block_locs = dict()
                for ap in attachable_polyominoes:
                    for block in attachable_polyominoes[ap].blocks:
                        if block in block_locs:
                            SimUtils.globalApp.OutputMessage(f'Violation of local determinism in at least one location. Two polyominoes could have been placed with blocks in location {block}:')
                            SimUtils.globalApp.OutputMessage(f'\t#1: {block_locs[block]}')
                            SimUtils.globalApp.OutputMessage(f'\t#2: {ap}')
                        else:
                            block_locs[block] = ap
                
            added_polyomino_hash = random.sample(attachable_polyominoes.keys(), 1)[0]
            added_polyomino = attachable_polyominoes[added_polyomino_hash]
            if len(attachable_polyominoes) > 1:
                if SimUtils.globalApp.canvas.refreshRate == 0:
                    SimUtils.globalApp.OutputMessage(f'One attachment of possible {len(attachable_polyominoes)} was randomly selected.')
            self.AddPolyomino(added_polyomino)
            return [added_polyomino]
        else:
            SimUtils.globalApp.OutputMessage(f"No polyomino attachments possible")
            return None

    def WriteToPDForSVG(self, pdfFileName="", svgFileName="", asSequence=False, displayMask=None, sideLen=None, z_range=None):
    
        if self.numTiles < 1:
            return False
        
        pdfCanvas = None
        svgCanvas = None
        
        if displayMask is None:
            displayMask = DISPLAY_ALL
    
        self.setDimensions()
        width = self.width
        height = self.height
        CELLSIZE = sideLen
    
        #(pageWidth, pageHeight) = (width*pdfTileDim+width*2*pdfTabDim, height*pdfTileDim+height*2*pdfTabDim)
        (pageWidth, pageHeight) = (width*CELLSIZE, height*CELLSIZE)
        #print("(w,h): " + str((pageWidth, pageHeight)))

        if len(pdfFileName) > 0:
            if os.path.isabs(pdfFileName) == False:
                pdfFileName = "./" + pdfFileName
                pdfFileName = os.path.abspath(pdfFileName)

            #c = canvas.Canvas(pdfFileName, pagesize=(width*pdfTileDim+width*2*pdfTabDim, height*pdfTileDim+height*2*pdfTabDim))
            pdfCanvas = canvas.Canvas(pdfFileName, pagesize=(pageWidth, pageHeight))
        elif len(svgFileName) > 0:
            if os.path.isabs(svgFileName) == False:
                svgFileName = "./" + svgFileName
                svgFileName = os.path.abspath(svgFileName)
            svgCanvas = svgwrite.Drawing(filename=svgFileName, debug=True)
            svgCanvas.viewbox(width=pageWidth, height=pageHeight)
        else:
            print("Error: No output file specified to WriteToPDForSVG")
            return False

        #from reportlab.lib.units import inch
        #pdfCanvas.translate(inch,inch)
        #MJP TODO: handle this case!!! - but only for PDF output?
        if asSequence == True:
            print('ERROR: Not implemented!')
            #===================================================================
            # for j in range(0,i+1):
            #     pt = (assembly[j][1][0]-minX, assembly[j][1][1]-minY)
            #     if isinstance(assembly[j][0], TileType):
            #         TileToPDF(pdfCanvas, assembly[j][0], pt, showLabels=True)
            #     elif j == i:
            #         TextToPDF(pdfCanvas, assembly[j][0], pt)
            # pdfCanvas.showPage()
            #===================================================================
        else:
            if (z_range is None):
                z_range = range(self.minZ, self.maxZ+1)
            else:
                z_range = range(z_range, z_range+1)
            drawn_polyominoes = set()
            for z_coord in z_range:
                for x_coord in range(self.minX, self.maxX+1):
                    for y_coord in range(self.minY, self.maxY+1):
                        if (x_coord, y_coord, z_coord) in self.locationMap:
                            polyomino = self.locationMap[(x_coord, y_coord, z_coord)]
                            ref_pt = polyomino.get_reference_point_loc()
                            ref_pt_str = str(ref_pt)
                            if not ref_pt_str in drawn_polyominoes:
                                
                                pt = (ref_pt[0]-self.minX, ref_pt[1]-self.minY)
                                trans = (pt[0]*CELLSIZE, pt[1]*CELLSIZE)
                                
                                if not pdfCanvas is None:
                                    pdfCanvas.translate(trans[0], trans[1])
                                polyomino.polyominoType.MakeImage(backgroundColor=None, mask=displayMask,
                                                    pdfCanvas=pdfCanvas, svgCanvas=svgCanvas,
                                                    trans=(trans[0], -trans[1]), sideLen=sideLen)
                                #polyomino.Draw(dc, (zero_x,zero_y), self.sideLen)
                                #polyomino.Draw2DOutline(dc, (zero_x,zero_y), self.sideLen)
                                if not pdfCanvas is None:
                                    pdfCanvas.translate(-trans[0], -trans[1])
                                drawn_polyominoes.add(ref_pt_str)
                                
        if not pdfCanvas is None:
            pdfCanvas.save()
        elif not svgCanvas is None:
            svgCanvas.save()
        return True