from PIL import Image
import json

img = Image.open('font.png')

with open('font_metadata.json') as md_file:
    meta = json.load(md_file)

width, height = img.size
fontSize = meta['size']

originOffsX = 0
originOffsY = 0
maxWidth = 0
maxHeight = 0
for c in meta['characters']:
    cd = meta['characters'][c]
    if originOffsX < cd['originX']:
        originOffsX = cd['originX']
    if originOffsY < cd['height'] - cd['originY']:
        originOffsY = cd['height'] - cd['originY']
print(originOffsX, originOffsY)

for c in meta['characters']:
    cd = meta['characters'][c]
    w = max([cd['width'], cd['width'] - cd['originX']])
    if maxWidth < w:
        maxWidth = w
    h = max([cd['height'], cd['originY']])
    if maxHeight < h:
        maxHeight = h
n = 1
row = (2 ** n) // maxWidth
col = (2 ** n) // maxHeight

while row * col < len(meta['characters']):
    n += 1
    row = (2 ** n) // maxWidth
    col = (2 ** n) // maxHeight

print(maxWidth, maxHeight, row, 2**n)

monoImg = Image.new('RGBA', (2 ** n, 2 ** n), 'black')

i = 0
char_arr = []
for char in meta['characters']:
    char_arr.append(char)
    nx = i % row
    ny = i // row

    px = nx * maxWidth
    py = ny * maxHeight

    c = meta['characters'][char]
    crop = (c['x'], c['y'], c['x'] + c['width'], c['y'] + c['height'])
    cropIm = img.crop(crop)

    newX = px + originOffsX - c['originX']
    newY = py + maxHeight - originOffsY + (c['height'] - c['originY']) - c['height']

    monoImg.paste(cropIm, (newX, newY))
    i += 1

monoMeta = {
    'width': maxWidth,
    'height': maxHeight,
    'size': 2 ** n,
    'characters': char_arr
}

with open('mono_font_metadata.json', 'w') as outfile:
    md = json.dump(monoMeta, outfile)

monoImg.save('mono_font.png')
monoImg.show()
