#version 130

#define CHARS 12

uniform ivec3 u_TileOffset;
uniform ivec3 u_Color;

uniform uint u_GlueLabel[6 * CHARS];
uniform uint u_GlueStrength;

uniform sampler2D u_FontTex;
const uint CHAR_WIDTH = 39u;
const uint CHAR_HEIGHT = 66u;

in vec4 v_Position;
in vec4 v_Normal;
in vec2 v_UV;
flat in int v_Face;

const float AMBIENT_CONTRIB = 0.8;
const float DIFFUSE_CONTRIB = 1.0 - AMBIENT_CONTRIB;

const vec4 LIGHT_POSITION = vec4(-5.0, 5.0, 0.0, 1.0);

out vec4  f_Color;
out ivec4 f_TileId;

void main()
{
    vec4 FragToLight = normalize(LIGHT_POSITION - v_Position);
    float diffuseAmount = max(0.0, dot(v_Normal, FragToLight));

    vec2 i_uv_f;
    vec2 mod_uv = modf(CHARS * v_UV, i_uv_f);
    ivec2 i_uv = ivec2(i_uv_f);

    vec4 color = vec4(u_Color / 255.0, 1.0);

    if (i_uv.y == CHARS / 2)
    {
        uint char_id = u_GlueLabel[v_Face * CHARS + i_uv.x];
        uint x = char_id % 26u;
        uint y = char_id / 26u;
        vec2 offset = vec2(float(x) / 26.0, float(y) / 15.0);
        vec2 uv = mod_uv * vec2(1.0 / 26.0, 1.0 / 15.0) + offset;
        //uv.y = 1 - uv.y;
        vec4 tex_color = texture(u_FontTex, uv);
        if (color.x*0.299 + color.y*0.587 + color.z*0.114 > 145.0 / 255.0)
        {
            color *= 1 - tex_color;
        }
        else
        {
            color = color * (1 - tex_color) + tex_color;
        }
        
    }
    // if (v_Face == 0)
    //     color = vec4(1.0, 0.0, 0.0, 1.0);
    // else if (v_Face == 1)
    //     color = vec4(0.0, 1.0, 0.0, 1.0);
    // else if (v_Face == 2)
    //     color = vec4(0.0, 0.0, 1.0, 1.0);
    // else if (v_Face == 3)
    //     color = vec4(0.0, 1.0, 1.0, 1.0);
    // else if (v_Face == 4)
    //     color = vec4(1.0, 0.0, 1.0, 1.0);
    // else
    //     color = vec4(1.0, 1.0, 0.0, 1.0);

    //f_Color = (vec4(vec3(u_Color), 255.0) / 255.0) * (diffuseAmount * DIFFUSE_CONTRIB + AMBIENT_CONTRIB);
    f_Color = color * (diffuseAmount * DIFFUSE_CONTRIB + AMBIENT_CONTRIB);
    f_TileId = ivec4(u_TileOffset, 0);
}